"""Tests for the summary module
"""

import datetime

import pytest

from sxolar.api.arxiv import Author, Category, Entry, Identifier
from sxolar.summary import Format, Section, Summary


class TestSection:
    """Test the summary module"""

    @pytest.fixture(autouse=True, scope="class")
    def results(self):
        """Return a list of results"""
        return [
            Entry.from_dict(
                {
                    "author": [
                        Author(name="Adrian Del Maestro", affiliation=None),
                        Author(name="Ian Affleck", affiliation=None),
                    ],
                    "category": [
                        Category(
                            term="cond-mat.stat-mech",
                            scheme="http://arxiv.org/schemas/atom",
                        )
                    ],
                    "id": Identifier(number="1005.5383", version="1", is_new=True),
                    "published": datetime.datetime(
                        2010, 5, 28, 20, 0, 5, tzinfo=datetime.timezone.utc
                    ),
                    "summary": "  Harmonically trapped ultra-cold atoms and helium-4 in nanopores provide new\nexperimental realizations of bosons in one dim...eement is obtained after\nincluding the leading irrelevant interactions in the Hamiltonian which are\ndetermined explicitly.\n",
                    "title": "Interacting bosons in one dimension and Luttinger liquid theory",
                    "updated": datetime.datetime(
                        2010, 5, 28, 20, 0, 5, tzinfo=datetime.timezone.utc
                    ),
                }
            ),
            Entry.from_dict(
                {
                    "author": [
                        Author(name="C. M. Herdman", affiliation=None),
                        Author(name="A. Rommal", affiliation=None),
                        Author(name="A. Del Maestro", affiliation=None),
                    ],
                    "category": [
                        Category(
                            term="cond-mat.stat-mech",
                            scheme="http://arxiv.org/schemas/atom",
                        ),
                        Category(
                            term="cond-mat.other",
                            scheme="http://arxiv.org/schemas/atom",
                        ),
                    ],
                    "id": Identifier(number="1312.6177", version="1", is_new=True),
                    "published": datetime.datetime(
                        2013, 12, 20, 23, 48, 25, tzinfo=datetime.timezone.utc
                    ),
                    "summary": "  A path integral Monte Carlo method based on the worm algorithm has been\ndeveloped to compute the chemical potential of int...re. We speculate on future applications of the proposed\ntechnique, including its use in studies of confined quantum fluids.\n",
                    "title": "Quantum Monte Carlo measurement of the chemical potential of helium-4",
                    "updated": datetime.datetime(
                        2013, 12, 20, 23, 48, 25, tzinfo=datetime.timezone.utc
                    ),
                }
            ),
        ]

    def test_init(self):
        """Test the summary class"""
        smy = Section("Test", None, [])
        assert isinstance(smy, Section)
        assert smy.name == "Test"
        assert smy.results == []
        assert smy.max_authors == 3
        assert smy.include_abstract is False

    def test_format_entry_plain(self, results):
        """Test the format_entry method"""
        smy = Section("TestPlain", query=None, results=results)
        entry = smy.results[0]
        entry_str = smy._format_entry(entry, Format.Plain)
        assert isinstance(entry_str, str)
        assert entry_str == (
            "Interacting bosons in one dimension and Luttinger liquid theory "
            "[1005.5383v1]\n"
            "Adrian Del Maestro, Ian Affleck\n"
            "\n"
            "http://arxiv.org/abs/1005.5383v1\n"
        )

    def test_format_entry_html(self, results):
        """Test the format_entry method"""
        smy = Section("TestPlain", query=None, results=results)
        entry = smy.results[0]
        entry_str = smy._format_entry(entry, Format.Html)
        assert isinstance(entry_str, str)
        assert entry_str == (
            '<p><h3><a href="http://arxiv.org/abs/1005.5383v1">Interacting bosons in one '
            "dimension and Luttinger liquid theory [1005.5383v1]</a></h3><br>Adrian Del "
            "Maestro, Ian Affleck<br><br></p>"
        )

    def test_to_text(self, results):
        """Test the to_text method"""
        smy = Section("TestPlain", query=None, results=results)
        text = smy.to_text()
        assert isinstance(text, str)
        assert text == (
            "TestPlain:\n"
            "Interacting bosons in one dimension and Luttinger liquid theory "
            "[1005.5383v1]\n"
            "Adrian Del Maestro, Ian Affleck\n"
            "\n"
            "http://arxiv.org/abs/1005.5383v1\n"
            "\n"
            "Quantum Monte Carlo measurement of the chemical potential of helium-4 "
            "[1312.6177v1]\n"
            "C. M. Herdman, A. Rommal, A. Del Maestro\n"
            "\n"
            "http://arxiv.org/abs/1312.6177v1\n"
        )

    def test_to_html(self, results):
        """Test the to_html method"""
        smy = Section("TestHtml", query=None, results=results)
        html = smy.to_html()
        assert isinstance(html, str)
        assert html == (
            "<h2>TestHtml</h2>\n"
            '<p><h3><a href="http://arxiv.org/abs/1005.5383v1">Interacting bosons in one '
            "dimension and Luttinger liquid theory [1005.5383v1]</a></h3><br>Adrian Del "
            "Maestro, Ian Affleck<br><br></p>\n"
            '<p><h3><a href="http://arxiv.org/abs/1312.6177v1">Quantum Monte Carlo '
            "measurement of the chemical potential of helium-4 "
            "[1312.6177v1]</a></h3><br>C. M. Herdman, A. Rommal, A. Del "
            "Maestro<br><br></p>"
        )

    def test_refresh(self):
        """Test the refresh method"""
        smy = Section.from_combo("Test1", authors=["Adrian Del Maestro"])
        smy.refresh()
        assert smy.results is not None
        assert len(smy.results) == 50
        assert isinstance(smy.results[0], Entry)
        assert smy.results[0].author == [
            Author(name="Adrian Del Maestro", affiliation=None),
            Author(name="Ian Affleck", affiliation=None),
        ]
        assert smy.results[0].category == [
            Category(
                term="cond-mat.stat-mech",
                scheme="http://arxiv.org/schemas/atom",
            )
        ]
        assert smy.results[0].id == Identifier(
            number="1005.5383", version="1", is_new=True
        )
        assert smy.results[0].published == datetime.datetime(
            2010, 5, 28, 20, 0, 5, tzinfo=datetime.timezone.utc
        )
        assert smy.results[0].summary == (
            "  Harmonically trapped ultra-cold atoms and helium-4 in nanopores provide "
            "new\n"
            "experimental realizations of bosons in one dimension, motivating the search "
            "for\n"
            "a more complete theoretical understanding of their low energy properties. "
            "Worm\n"
            "algorithm path integral quantum Monte Carlo results for interacting bosons\n"
            "restricted to the one dimensional continuum are compared to the finite\n"
            "temperature and system size predictions of Luttinger liquid theory. For "
            "large\n"
            "system sizes at low temperature, excellent agreement is obtained after\n"
            "including the leading irrelevant interactions in the Hamiltonian which are\n"
            "determined explicitly.\n"
        )
        assert (
            smy.results[0].title
            == "Interacting bosons in one dimension and Luttinger liquid theory"
        )
        assert smy.results[0].updated == datetime.datetime(
            2010, 5, 28, 20, 0, 5, tzinfo=datetime.timezone.utc
        )


class TestSummary:
    """Test the summary module"""

    @pytest.fixture(autouse=True, scope="class")
    def results(self):
        """Return a list of results"""
        return [
            Entry.from_dict(
                {
                    "author": [
                        Author(name="Adrian Del Maestro", affiliation=None),
                        Author(name="Ian Affleck", affiliation=None),
                    ],
                    "category": [
                        Category(
                            term="cond-mat.stat-mech",
                            scheme="http://arxiv.org/schemas/atom",
                        )
                    ],
                    "id": Identifier(number="1005.5383", version="1", is_new=True),
                    "published": datetime.datetime(
                        2010, 5, 28, 20, 0, 5, tzinfo=datetime.timezone.utc
                    ),
                    "summary": "  Harmonically trapped ultra-cold atoms and helium-4 in nanopores provide new\nexperimental realizations of bosons in one dim...eement is obtained after\nincluding the leading irrelevant interactions in the Hamiltonian which are\ndetermined explicitly.\n",
                    "title": "Interacting bosons in one dimension and Luttinger liquid theory",
                    "updated": datetime.datetime(
                        2010, 5, 28, 20, 0, 5, tzinfo=datetime.timezone.utc
                    ),
                }
            ),
            Entry.from_dict(
                {
                    "author": [
                        Author(name="C. M. Herdman", affiliation=None),
                        Author(name="A. Rommal", affiliation=None),
                        Author(name="A. Del Maestro", affiliation=None),
                    ],
                    "category": [
                        Category(
                            term="cond-mat.stat-mech",
                            scheme="http://arxiv.org/schemas/atom",
                        ),
                        Category(
                            term="cond-mat.other",
                            scheme="http://arxiv.org/schemas/atom",
                        ),
                    ],
                    "id": Identifier(number="1312.6177", version="1", is_new=True),
                    "published": datetime.datetime(
                        2013, 12, 20, 23, 48, 25, tzinfo=datetime.timezone.utc
                    ),
                    "summary": "  A path integral Monte Carlo method based on the worm algorithm has been\ndeveloped to compute the chemical potential of int...re. We speculate on future applications of the proposed\ntechnique, including its use in studies of confined quantum fluids.\n",
                    "title": "Quantum Monte Carlo measurement of the chemical potential of helium-4",
                    "updated": datetime.datetime(
                        2013, 12, 20, 23, 48, 25, tzinfo=datetime.timezone.utc
                    ),
                }
            ),
        ]

    def test_to_text(self, results):
        """Test the to_text method"""
        sec1 = Section("TestSec1", query=None, results=results[:1])
        sec2 = Section("TestSec2", query=None, results=results[1:])
        smy = Summary("Test", [sec1, sec2])
        text = smy.to_text()
        assert isinstance(text, str)
        assert text == (
            "TestSec1:\n"
            "Interacting bosons in one dimension and Luttinger liquid theory "
            "[1005.5383v1]\n"
            "Adrian Del Maestro, Ian Affleck\n"
            "\n"
            "http://arxiv.org/abs/1005.5383v1\n"
            "\n"
            "\n"
            "TestSec2:\n"
            "Quantum Monte Carlo measurement of the chemical potential of helium-4 "
            "[1312.6177v1]\n"
            "C. M. Herdman, A. Rommal, A. Del Maestro\n"
            "\n"
            "http://arxiv.org/abs/1312.6177v1\n"
        )

    def test_to_html(self, results):
        """Test the to_text method"""
        sec1 = Section("TestSec1", query=None, results=results[:1])
        sec2 = Section("TestSec2", query=None, results=results[1:])
        smy = Summary("Test", [sec1, sec2])
        text = smy.to_html()
        assert isinstance(text, str)
        assert text == (
            "<h2>TestSec1</h2>\n"
            '<p><h3><a href="http://arxiv.org/abs/1005.5383v1">Interacting bosons in one '
            "dimension and Luttinger liquid theory [1005.5383v1]</a></h3><br>Adrian Del "
            "Maestro, Ian Affleck<br><br></p><br><br><h2>TestSec2</h2>\n"
            '<p><h3><a href="http://arxiv.org/abs/1312.6177v1">Quantum Monte Carlo '
            "measurement of the chemical potential of helium-4 "
            "[1312.6177v1]</a></h3><br>C. M. Herdman, A. Rommal, A. Del "
            "Maestro<br><br></p>"
        )

    def test_refresh(self):
        """Test the refresh method"""
        sec = Section.from_combo("Test1", authors=["Adrian Del Maestro"])
        smy = Summary("Test", [sec])
        smy.refresh()
        smy = smy.sections[0]
        assert smy.results is not None
        assert len(smy.results) == 50
        assert isinstance(smy.results[0], Entry)
        assert smy.results[0].author == [
            Author(name="Adrian Del Maestro", affiliation=None),
            Author(name="Ian Affleck", affiliation=None),
        ]
        assert smy.results[0].category == [
            Category(
                term="cond-mat.stat-mech",
                scheme="http://arxiv.org/schemas/atom",
            )
        ]
        assert smy.results[0].id == Identifier(
            number="1005.5383", version="1", is_new=True
        )
        assert smy.results[0].published == datetime.datetime(
            2010, 5, 28, 20, 0, 5, tzinfo=datetime.timezone.utc
        )
        assert smy.results[0].summary == (
            "  Harmonically trapped ultra-cold atoms and helium-4 in nanopores provide "
            "new\n"
            "experimental realizations of bosons in one dimension, motivating the search "
            "for\n"
            "a more complete theoretical understanding of their low energy properties. "
            "Worm\n"
            "algorithm path integral quantum Monte Carlo results for interacting bosons\n"
            "restricted to the one dimensional continuum are compared to the finite\n"
            "temperature and system size predictions of Luttinger liquid theory. For "
            "large\n"
            "system sizes at low temperature, excellent agreement is obtained after\n"
            "including the leading irrelevant interactions in the Hamiltonian which are\n"
            "determined explicitly.\n"
        )
        assert (
            smy.results[0].title
            == "Interacting bosons in one dimension and Luttinger liquid theory"
        )
        assert smy.results[0].updated == datetime.datetime(
            2010, 5, 28, 20, 0, 5, tzinfo=datetime.timezone.utc
        )



"""Utilities for reading and writing configuration files for sxolar

Useful for storing user preferences and settings for saved queries
and configured summaries.

Sample Config file:

The example below defines 2 summaries, each with a list of sections.
Each section has a name, a list of authors, a list of topics, and a trailing
time period.

```yaml
summary name 1:
  - name: "Section 1: Topic A x Authors 1, 2 | Recent 2 Weeks"
    authors: ["Author 1", "Author 2"]
    alls: ["Topic A"]
    trailing:
      num: 14
      unit: "days"

  - name: "Section 2: Topic B x Authors 3, 4 | Recent 2 Months"
    authors: ["Author 3", "Author 4"]
    alls: ["Topic B"]
    trailing:
      num: 2
      unit: "months"

summary name 2:
  - name: "Section 1: Topic C x Authors 5, 6"
    authors: ["Author 5", "Author 6"]
    alls: ["Topic C"]
```

"""

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Union

from ruamel.yaml import YAML

from sxolar.summary import Section, Summary


@dataclass
class Config:
    """Configuration settings for sxolar

    Args:
        named_queries: A dictionary of named queries
    """

    info: dict
    summaries: Dict[str, Summary] = None

    def __post_init__(self):
        """Initialize the configuration settings"""
        # Parse the summaries from the configuration settings
        self.summaries = {}
        for name, sections in self.info.items():
            # Create the sections for the summary
            _sections = []
            for section in sections:
                _sections.append(Section.from_combo(**section))

            # Create the summary
            self.summaries[name] = Summary(name=name, sections=_sections)

    @staticmethod
    def load(path: Union[str, Path]):
        """Load a configuration file from a given path

        Args:
            path: str, the path to the configuration file

        Returns:
            Config, the configuration settings
        """
        yaml = YAML(typ="safe", pure=True)
        with open(path, "r") as f:
            data = yaml.load(f)
        return Config(info=data)



"""Higher-level api for searching for papers, uses an object interface
with overridden magic methods for syntactic sugar
"""

import datetime
from typing import Iterable, List, Union

from sxolar.api import arxiv
from sxolar.api.arxiv import (
    FIELD_ENTRY_UPDATED,
    LogicalOperator,
    SearchField,
    SortBy,
    SortOrder,
)


class Query:
    """Represents a query clause for the arxiv API

    Attributes:
        value (str): The value to search for
        operator (str): The operator to use
    """

    def __init__(
        self,
        value: str,
        filter_authors: List[str] = None,
    ):
        """Creates a new query

        Args:
            value:
                str, the value to search for
            filter_authors:
                List[str], the authors to filter by
        """
        self.value = value
        self.filter_authors = filter_authors or []

    def __str__(self):
        """Returns the string representation of the query"""
        return self.value

    def __and__(self, other):
        """Overloads the and operator to create a new query"""
        return self.and_(other)

    def __or__(self, other):
        """Overloads the or operator to create a new query"""
        return self.or_(other)

    def __sub__(self, other):
        """Overloads the subtraction operator to create a new query"""
        return self.and_not(other)

    def and_(self, other: Union[str, "Query"]):
        """Join two queries with the AND operator

        Args:
            other:
                str, the other query to join with

        Returns:
            Query: A new query object
        """
        return Query(
            f"{self}{LogicalOperator.AND}{other}",
            filter_authors=list(set(self.filter_authors) & set(other.filter_authors)),
        )

    def and_not(self, other):
        """Join two queries with the AND NOT operator

        Args:
            other:
                str, the other query to join with

        Returns:
            Query: A new query object
        """
        return Query(
            f"{self}{LogicalOperator.AND_NOT}{other}",
            filter_authors=list(set(self.filter_authors) - set(other.filter_authors)),
        )

    def or_(self, other):
        """Join two queries with the OR operator

        Args:
            other:
                str, the other query to join with

        Returns:
            Query: A new query object
        """
        return Query(
            f"{self}{LogicalOperator.OR}{other}",
            filter_authors=list(set(self.filter_authors) | set(other.filter_authors)),
        )

    def join(
        self, *others: Iterable["Query"], operator: LogicalOperator = LogicalOperator.OR
    ):
        """Join multiple queries with an operator

        Args:
            others:
                Iterable[Query], the queries to join
            operator:
                LogicalOperator, the operator to use to join the queries

        Returns:
            Query: A new query object
        """
        if not others:
            return self

        value = self.value
        authors = set(self.filter_authors)
        for other in others:
            value = f"{value}{operator}{other}"
            authors |= set(other.filter_authors)

        return Query(value, filter_authors=list(sorted(authors))).wrap()

    def wrap(self):
        """Wrap the query in parenthesis

        Returns:
            Query: A new query object
        """
        return Query(f"({self})", filter_authors=self.filter_authors)

    def search(
        self,
        start: int = 0,
        max_results: int = 10,
        sort_by: SortBy = SortBy.Relevance,
        sort_order: SortOrder = SortOrder.Descending,
        min_date: datetime.datetime = None,
        max_date: datetime.datetime = None,
        date_filter_field: str = FIELD_ENTRY_UPDATED,
    ):
        """Searches the arxiv API with the query

        Args:
            start:
                int, optional, The starting index of the results
            max_results:
                int, optional, The maximum number of results to return

        Returns:
            list: A list of dictionaries representing the search results
        """
        results = arxiv._query(
            self.value,
            id_list=None,
            start=start,
            max_results=max_results,
            sort_by=sort_by,
            sort_order=sort_order,
            min_date=min_date,
            max_date=max_date,
            date_filter_field=date_filter_field,
        )

        # Apply filter authors if any
        if self.filter_authors:
            results = [
                entry for entry in results if entry.filter_authors(self.filter_authors)
            ]

        return results

    def to_str(self) -> str:
        """Returns the string representation of the query"""
        return self.value

    @staticmethod
    def from_str(value: str):
        """Creates a new query from a string

        Args:
            value:
                str, the value to search for
        """
        return Query(value)

    @staticmethod
    def from_authors(*authors: Iterable[str]):
        """Creates a new author query

        Args:
            authors:
                str, the name of the author, "First Last"
        """
        authors = [Author(author) for author in authors]
        query = authors[0]
        if authors[1:]:
            query = query.join(*authors[1:], operator=LogicalOperator.OR)
        return query

    @staticmethod
    def from_titles(*titles: Iterable[str]):
        """Creates a new title query

        Args:
            titles:
                str, the title of the paper
        """
        titles = [Title(title) for title in titles]
        query = titles[0]
        if titles[1:]:
            query = query.join(*titles[1:], operator=LogicalOperator.OR)
        return query

    @staticmethod
    def from_abstracts(*abstracts: Iterable[str]):
        """Creates a new abstract query

        Args:
            abstracts:
                str, the abstract of the paper
        """
        abstracts = [Abstract(abstract) for abstract in abstracts]
        query = abstracts[0]
        if abstracts[1:]:
            query = query.join(*abstracts[1:], operator=LogicalOperator.OR)
        return query

    @staticmethod
    def from_alls(*alls: Iterable[str], operator: LogicalOperator = LogicalOperator.OR):
        """Creates a new all query

        Args:
            alls:
                str, the value to search for
        """
        alls = [All(all_) for all_ in alls]
        query = alls[0]
        if alls[1:]:
            query = query.join(*alls[1:], operator=operator)
        return query

    @staticmethod
    def from_journal_refs(*journal_refs: Iterable[str]):
        """Creates a new journal reference query

        Args:
            journal_refs:
                str, the journal reference
        """
        journal_refs = [JournalRef(journal_ref) for journal_ref in journal_refs]
        query = journal_refs[0]
        if journal_refs[1:]:
            query = query.join(*journal_refs[1:], operator=LogicalOperator.OR)
        return query

    @staticmethod
    def from_categories(*categories: Iterable[str]):
        """Creates a new category query

        Args:
            categories:
                str, the category
        """
        categories = [Category(category) for category in categories]
        query = categories[0]
        if categories[1:]:
            query = query.join(*categories[1:], operator=LogicalOperator.OR)
        return query

    @staticmethod
    def from_combo(
        authors: Iterable[str] = None,
        titles: Iterable[str] = None,
        abstracts: Iterable[str] = None,
        alls: Iterable[str] = None,
        journal_refs: Iterable[str] = None,
        categories: Iterable[str] = None,
        operator: LogicalOperator = LogicalOperator.AND,
        filter_authors: bool = False,
        alls_operator: LogicalOperator = LogicalOperator.OR,
    ):
        """Creates a new combo query

        Args:
            authors:
                Iterable[str], the name of the author, "First Last"
            titles:
                Iterable[str], the title of the paper
            abstracts:
                Iterable[str], the abstract of the paper
            alls:
                Iterable[str], the value to search for
            journal_refs:
                Iterable[str], the journal reference
            categories:
                Iterable[str], the category
        """
        queries = []
        if authors:
            queries.append(Query.from_authors(*authors))
        if titles:
            queries.append(Query.from_titles(*titles))
        if abstracts:
            queries.append(Query.from_abstracts(*abstracts))
        if alls:
            queries.append(Query.from_alls(*alls, operator=alls_operator))
        if journal_refs:
            queries.append(Query.from_journal_refs(*journal_refs))
        if categories:
            queries.append(Query.from_categories(*categories))

        query = queries[0]
        if queries[1:]:
            query = query.join(*queries[1:], operator=operator)

        # Add author filters if needed
        if filter_authors and authors:
            query.filter_authors = list(set(authors))

        return query


class Author(Query):
    """Represents an author query for the arxiv API"""

    def __init__(self, name: str):
        """Creates a new author query

        Args:
            name:
                str, the name of the author, "First Last"
        """
        if not name.startswith(SearchField.AUTHOR):
            name = f"{SearchField.AUTHOR}:{name}"
        super().__init__(name)


class Title(Query):
    """Represents a title query for the arxiv API"""

    def __init__(self, title: str):
        """Creates a new title query

        Args:
            title:
                str, the title of the paper
        """
        if not title.startswith(SearchField.TITLE):
            title = f"{SearchField.TITLE}:{title}"
        super().__init__(title)


class Abstract(Query):
    """Represents an abstract query for the arxiv API"""

    def __init__(self, abstract: str):
        """Creates a new abstract query

        Args:
            abstract:
                str, the abstract of the paper
        """
        if not abstract.startswith(SearchField.ABSTRACT):
            abstract = f"{SearchField.ABSTRACT}:{abstract}"
        super().__init__(abstract)


class All(Query):
    """Represents an all query for the arxiv API"""

    def __init__(self, all_: str):
        """Creates a new all query

        Args:
            all_:
                str, the value to search for
        """
        if not all_.startswith(SearchField.ALL):
            all_ = f"{SearchField.ALL}:{all_}"
        super().__init__(all_)


class JournalRef(Query):
    """Represents a journal reference query for the arxiv API"""

    def __init__(self, journal_ref: str):
        """Creates a new journal reference query

        Args:
            journal_ref:
                str, the journal reference
        """
        if not journal_ref.startswith(SearchField.JOURNAL_REFERENCE):
            journal_ref = f"{SearchField.JOURNAL_REFERENCE}:{journal_ref}"
        super().__init__(journal_ref)


class Category(Query):
    """Represents a category query for the arxiv API"""

    def __init__(self, category: str):
        """Creates a new category query

        Args:
            category:
        """
        if not category.startswith(SearchField.CATEGORY):
            category = f"{SearchField.CATEGORY}:{category}"
        super().__init__(category)



"""Sxolar command line interface.

Current commands relate solely to running queries based on config files,
primarily used for generating summaries of arXiv papers and sending emails
with the results.
"""

import argparse
from argparse import Namespace
from typing import List

from sxolar import Query
from sxolar.api.arxiv import SortBy, SortOrder
from sxolar.config import Config
from sxolar.summary import Section
from sxolar.util import gmail


def parse_args(args: List[str] = None) -> Namespace:
    """Parse command line arguments for the sxolar command line interface.

    Args:
        args: List[str], the list of command line arguments

    Returns:
        Namespace, the parsed command line arguments
    """
    parser = argparse.ArgumentParser(
        description="Command line interface for the sxolar package"
    )

    # Add subparsers for different commands
    subparsers = parser.add_subparsers(dest="command")

    # Add the query parser
    add_query_parser(subparsers=subparsers)

    # Add the summary parser
    add_summary_parser(subparsers=subparsers)

    # Parse args
    return parser.parse_args(args)


def add_query_parser(subparsers: argparse._SubParsersAction) -> None:
    """Add a parser for the query command to the main parser."""

    # Parser for the query command
    query_parser = subparsers.add_parser("query", help="Run a query")
    query_parser.add_argument(
        "--value",
        type=str,
        help="Fully formatted query string, e.g. 'au:del maestro'. Optional. If "
        "specified, this will override the query specified in the configuration "
        "or in subsequent arguments.",
        required=False,
    )
    # Argument for list of authors
    query_parser.add_argument(
        "--authors",
        type=str,
        nargs="+",
        help="List of authors to search for. Optional. If specified, this will "
        "override the authors specified in the configuration or in subsequent "
        "arguments.",
        required=False,
    )
    # Argument for list of titles
    query_parser.add_argument(
        "--titles",
        type=str,
        nargs="+",
        help="List of titles to search for. Optional. If specified, this will "
        "override the titles specified in the configuration or in subsequent "
        "arguments.",
        required=False,
    )
    # Argument for list of abstracts
    query_parser.add_argument(
        "--abstracts",
        type=str,
        nargs="+",
        help="List of abstracts to search for. Optional. If specified, this will "
        "override the abstracts specified in the configuration or in subsequent "
        "arguments.",
        required=False,
    )
    # Argument for list of categories
    query_parser.add_argument(
        "--categories",
        type=str,
        nargs="+",
        help="List of categories to search for. Optional. If specified, this will "
        "override the categories specified in the configuration or in subsequent "
        "arguments.",
        required=False,
    )
    # Argument for list of topics
    query_parser.add_argument(
        "--alls",
        type=str,
        nargs="+",
        help="List of alls to search for. Optional. If specified, this will "
        "override the topics specified in the configuration or in subsequent "
        "arguments.",
        required=False,
    )
    # Argument for sort by
    query_parser.add_argument(
        "--sort-by",
        type=str,
        help="Sort by field. Optional. If specified, this will override the sort "
        "field specified in the configuration or in subsequent arguments.",
        required=False,
    )
    # Argument for max results
    query_parser.add_argument(
        "--max-results",
        type=int,
        help="Maximum number of results to return. Optional. If specified, this will "
        "override the maximum number of results specified in the configuration or "
        "in subsequent arguments.",
        required=False,
    )
    # Time filtering arguments
    query_parser.add_argument(
        "--trailing",
        type=int,
        help="Number of units to include in the trailing period. Optional. If "
        "specified, this will override the trailing period specified in the "
        "configuration or in subsequent arguments.",
        required=False,
    )


def add_summary_parser(subparsers: argparse._SubParsersAction) -> None:
    """Add a parser for the summary command to the main parser."""

    # Parser for the summary command
    summary_parser = subparsers.add_parser("summary", help="Generate a summary")
    summary_parser.add_argument(
        "--name",
        type=str,
        help="Name of the summary to generate. Required.",
        required=True,
    )
    summary_parser.add_argument(
        "--output",
        type=str,
        choices=["print", "email"],
        help="Output file for the summary. Optional. If specified, this will override "
        "the output file specified in the configuration or in subsequent arguments.",
        default="print",
        required=False,
    )
    # Add argument for email from
    summary_parser.add_argument(
        "--email-from",
        type=str,
        help="Email address to send the summary from. Optional. Only required if "
        "sending the summary by email.",
        default="chalkdust@sxolar.org",
        required=False,
    )
    # Add argument for email to
    summary_parser.add_argument(
        "--email-to",
        type=str,
        nargs="+",
        help="Email address to send the summary to. Optional. Only required if "
        "sending the summary by email.",
        required=False,
    )
    # Add argument for email subject
    summary_parser.add_argument(
        "--email-subject",
        type=str,
        help="Email subject for the summary. Optional. Only required if sending the "
        "summary by email.",
        default="ArXiv Summary from sXolar",
        required=False,
    )
    # Add argument for gmail app password
    summary_parser.add_argument(
        "--gmail-app-password",
        type=str,
        help="Gmail app password for sending the email. Optional. Only required if "
        "sending the summary by email.",
        required=False,
    )
    # Add argument for config file
    summary_parser.add_argument(
        "--config",
        type=str,
        help="Path to the configuration file. Optional. If specified, this will "
        "override the configuration file specified in the configuration or in "
        "subsequent arguments.",
        required=False,
    )


def query(
    authors: List[str] = None,
    titles: List[str] = None,
    abstracts: List[str] = None,
    categories: List[str] = None,
    alls: List[str] = None,
    sort_by: SortBy = None,
    sort_order: SortOrder = None,
    max_results: int = None,
    trailing: int = None,
    value: str = None,
) -> None:
    """Run a query based on the command line arguments."""
    # Build query object
    if value is not None:
        query = Query(value=value)
    else:
        query = Query.from_combo(
            authors=authors,
            titles=titles,
            abstracts=abstracts,
            categories=categories,
            alls=alls,
        )

    section = Section(
        name="Query Results",
        query=query,
        max_results=max_results,
        trailing=trailing,
        sort_by=sort_by,
        sort_order=sort_order,
    )

    # Run the query
    section.refresh()

    # Print the results
    print(section.to_text())


def summary(
    config_path: str,
    name: str,
    output: str = "print",
    email_from: str = "chalkdust@sxolar.org",
    email_to: List[str] = None,
    email_subject: str = "ArXiv Summary from sXolar",
    gmail_app_password: str = None,
) -> None:
    """Generate a summary based on the command line arguments."""
    # Load config
    try:
        config = Config.load(config_path)
    except Exception as e:
        raise ValueError(f"Error loading configuration file: {e}") from e

    # Get the summary associated to the name
    summary = config.summaries.get(name, None)
    if summary is None:
        raise ValueError(
            f"Summary not found: {name}, options are {list(config.summaries.keys())}"
        )

    # Refresh the summary
    summary.refresh()

    # Print the summary if that is the output
    if output == "print":
        print(summary.to_text())

    # Send the summary by email if that is the output
    elif output == "email":
        if email_to is None:
            raise ValueError("Email to address must be specified for email output")
        if gmail_app_password is None:
            # Check env var
            if gmail.EMAIL_APP_PASSWORD is None:
                raise ValueError("Gmail app password must be specified for email "
                                 "output, can be set as an environment variable "
                                 "$SXOLAR_EMAIL_APP_PASSWORD or specify it as the")
        body = summary.to_html()
        # Send email
        gmail.send_email(
            subject=email_subject,
            to=email_to,
            body=body,
            from_email=email_from,
            app_password=gmail_app_password,
            is_plain=False,
        )

    # Invalid output
    else:
        raise ValueError(f"Invalid output: {output}, options are print, email")


def main():
    """Main function for the sxolar command line interface."""
    args = parse_args()

    # Dispatch for different commands
    # Query command
    if args.command == "query":
        # Convert the sort by argument to a SortBy enum
        if args.sort_by is None:
            sort_by = None
        else:
            sort_by = {
                "relevance": SortBy.Relevance,
                "lastUpdatedDate": SortBy.LastUpdatedDate,
                "submittedDate": SortBy.SubmittedDate,
            }.get(args.sort_by, None)
            if sort_by is None:
                raise ValueError(
                    f"Invalid sort by argument: {args.sort_by}, options "
                    f"are relevance, lastUpdatedDate, submittedDate"
                )

        query(
            authors=args.authors,
            titles=args.titles,
            abstracts=args.abstracts,
            categories=args.categories,
            alls=args.alls,
            sort_by=sort_by,
            max_results=args.max_results,
            trailing=args.trailing,
            value=args.value,
        )

    # Summary command
    elif args.command == "summary":
        summary(
            config_path=args.config,
            name=args.name,
            output=args.output,
            email_from=args.email_from,
            email_to=args.email_to,
            email_subject=args.email_subject,
            gmail_app_password=args.gmail_app_password,
        )

    # Invalid command
    else:
        raise ValueError(f"Invalid command: {args.command}")


if __name__ == "__main__":
    main()



"""Tests for gmail module"""

import pytest

from sxolar.util import gmail

class TestGmail:
    """Test the gmail module"""

    @pytest.mark.skip("Something about smtp is broken in pytest")
    def test_send_email(self):
        """Test the send_email function"""
        body = (
            "<h2>TestHtml</h2>\n"
            '<p><h3><a href="http://arxiv.org/abs/1005.5383v1">Interacting bosons in one '
            "dimension and Luttinger liquid theory [1005.5383v1]</a></h3><br>Adrian Del "
            "Maestro, Ian Affleck<br><br></p>\n"
            '<p><h3><a href="http://arxiv.org/abs/1312.6177v1">Quantum Monte Carlo '
            "measurement of the chemical potential of helium-4 "
            "[1312.6177v1]</a></h3><br>C. M. Herdman, A. Rommal, A. Del "
            "Maestro<br><br></p>"
        )
        gmail.send_email(
            subject="Test Email Sxolar",
            to="jameswkennington@gmail.com",
            body=body,
            is_plain=False,
        )



"""Tests for config utilities
"""

import pathlib

from sxolar.config import Config

PATH_TESTS = pathlib.Path(__file__).parent


class TestConfig:
    """Test group for config utilities"""

    def test_load(self):
        """Test loading a configuration file"""
        # Given a path to a configuration file
        path = PATH_TESTS / "data" / "sample.yml"

        # When loading the configuration file
        config = Config.load(path)

        # Then the configuration settings should be loaded
        assert config.info == {
            "summary name 1": [
                {
                    "name": "Section 1: Topic A x Authors 1, 2 | Recent 2 Weeks",
                    "authors": ["Author 1", "Author 2"],
                    "alls": ["Topic A"],
                    "trailing": {"num": 14, "unit": "days"},
                },
                {
                    "name": "Section 2: Topic B x Authors 3, 4 | Recent 2 Months",
                    "authors": ["Author 3", "Author 4"],
                    "alls": ["Topic B"],
                    "trailing": {"num": 2, "unit": "months"},
                },
            ],
            "summary name 2": [
                {
                    "name": "Section 1: Topic C x Authors 5, 6",
                    "authors": ["Author 5", "Author 6"],
                    "alls": ["Topic C"],
                },
            ],
        }

    def test_summaries(self):
        """Test loading a configuration file"""
        # Given a path to a configuration file
        path = PATH_TESTS / "data" / "sample.yml"

        # When loading the configuration file
        config = Config.load(path)

        assert len(config.summaries) == 2
        assert list(sorted(config.summaries.keys())) == [
            "summary name 1",
            "summary name 2",
        ]

        # Check details of summary objects
        summary = config.summaries["summary name 1"]
        assert summary.name == "summary name 1"
        assert len(summary.sections) == 2
        assert (
            summary.sections[0].name
            == "Section 1: Topic A x Authors 1, 2 | Recent 2 Weeks"
        )
        assert str(summary.sections[0].query) == (
            "((au:Author 1 OR au:Author 2) AND all:Topic A)"
        )



"""Tests for the arxiv module.
"""

import datetime

from sxolar.api import arxiv


class TestArxivAPI:
    """Test the arxiv API"""

    def test_query(self):
        """Test the query function,

        This test mimicks the behavior of the API documentation, which gives "3" results
            Docs Test Query: "http://export.arxiv.org/api/query?search_query=au:del_maestro+ANDNOT+%28ti:checkerboard+OR+ti:Pyrochlore%29"
        """
        res = arxiv.query(author="del maestro")
        assert len(res) == 10

    def test_query_last_updated(self):
        """Test the query function,

        This test mimicks the behavior of the API documentation, which gives "3" results
            Docs Test Query: "http://export.arxiv.org/api/query?search_query=au:del_maestro+ANDNOT+%28ti:checkerboard+OR+ti:Pyrochlore%29"
        """
        # Make date filters for testing UTC
        date_end = datetime.datetime.now(datetime.UTC)
        date_start = date_end - datetime.timedelta(days=7)

        # Test the query
        res = arxiv.query(
            author="del maestro",
            sort_by=arxiv.SortBy.LastUpdatedDate,
            sort_order=arxiv.SortOrder.Descending,
            min_date=date_start,
            max_date=date_end,
        )

        # Check there is at least one result
        assert len(res) > 0

        # Check date filter worked
        for entry in res:
            assert date_start <= entry.updated <= date_end



"""Arxiv API wrappers for sxolar, low-level functions for querying the Arxiv API.
For more user-friendly functions, see the `sxolar.api.query` module.

References:
    [1] API Basics: https://info.arxiv.org/help/api/basics.html
    [2] Rate Limits: https://info.arxiv.org/help/api/tou.html
    [3] Search Query Language: https://info.arxiv.org/help/api/user-manual.html
        #query_details
    [4] Entry output format: https://info.arxiv.org/help/api/user-manual.html
        #_entry_metadata
    [5] ArXiv identifier format: https://info.arxiv.org/help/arxiv_identifier.html
"""

import collections
import datetime
import enum
import re
from dataclasses import dataclass
from typing import List, Union
from urllib import parse
from xml.etree import ElementTree

from defusedxml import ElementTree as SecureElementTree
from requests_ratelimiter import LimiterSession

# Impose the public ratelimit by default [2]
SESSION = LimiterSession(per_minute=20)

# Define the base URL for the Arxiv API [1]
URL_BASE = "http://export.arxiv.org/api/"
URL_QUERY = URL_BASE + "query"

# Define XML tags for the Arxiv API [4]
TAG_PREFIX = "{http://www.w3.org/2005/Atom}"
TAG_ENTRY = "entry"
TAG_TITLE = "title"
TAG_ID = "id"
TAG_PUBLISHED = "published"
TAG_UPDATED = "updated"
TAG_SUMMARY = "summary"
TAG_AUTHOR = "author"
TAG_NAME = "name"
TAG_AFFILIATION = "affiliation"
TAG_CATEGORY = "category"
TAG_TERM = "term"
TAG_SCHEME = "scheme"

# Identifier schema (NEW)
ID_PREFIX = "http://arxiv.org/abs/"
ID_NUM_PATTERN_NEW = r"[0-9]{4}\.[0-9]{4,5}"
ID_VERSION_PATTERN_NEW = "[0-9]+"
# ID Pattern with named groups for the number and version
ID_PATTERN_NEW = (
    f"(?P<number>{ID_NUM_PATTERN_NEW})(?:v(?P<version" f">{ID_VERSION_PATTERN_NEW}))?"
)

# Identifier schema (OLD) example: math.GT/0309136
ID_NUM_PATTERN = "[a-z-]+/[0-9]{6,8}"
ID_PATTERN_OLD = f"(?P<number>{ID_NUM_PATTERN})"

# Define author fields
FIELD_AUTHOR_NAME = "name"
FIELD_AUTHOR_AFFILIATION = "affiliation"
FIELDS_AUTHOR = (
    FIELD_AUTHOR_NAME,
    FIELD_AUTHOR_AFFILIATION,
)

# Define category fields
FIELD_CATEGORY_TERM = "term"
FIELD_CATEGORY_SCHEME = "scheme"
FIELDS_CATEGORY = (
    FIELD_CATEGORY_TERM,
    FIELD_CATEGORY_SCHEME,
)

# Define entry fields
FIELD_ENTRY_TITLE = "title"
FIELD_ENTRY_ID = "id"
FIELD_ENTRY_PUBLISHED = "published"
FIELD_ENTRY_UPDATED = "updated"
FIELD_ENTRY_SUMMARY = "summary"
FIELD_ENTRY_AUTHOR = "author"
FIELD_ENTRY_CATEGORY = "category"
FIELDS_ENTRY = (
    FIELD_ENTRY_TITLE,
    FIELD_ENTRY_ID,
    FIELD_ENTRY_PUBLISHED,
    FIELD_ENTRY_UPDATED,
    FIELD_ENTRY_SUMMARY,
    FIELD_ENTRY_AUTHOR,
    FIELD_ENTRY_CATEGORY,
)


Category = collections.namedtuple("Category", " ".join(FIELDS_CATEGORY))


@dataclass
class Author:
    """A dataclass for an author of an Arxiv entry.

    Args:
        name:
            str, the name of the author.
        affiliation:
            str, the affiliation of the author.
    """

    name: str
    affiliation: str

    def __str__(self) -> str:
        """Return the author as a string."""
        return self.name


@dataclass
class Identifier:
    """A dataclass for an Arxiv identifier.

    Args:
        number:
            str, the number of the identifier.
        version:
            str, the version of the identifier.
        is_new:
            bool, whether the identifier is in the new format.
    """

    number: str
    version: str
    is_new: bool

    def __str__(self) -> str:
        """Return the identifier as a string."""
        if self.is_new:
            return f"{self.number}v{self.version}" if self.version else self.number
        return self.number

    def link(self) -> str:
        """Formatted arxiv link for the identifier"""
        return f"{ID_PREFIX}{self}"


def parse_identifier(id_text: str) -> Identifier:
    """Parse an Arxiv identifier into its components.

    Args:
        id_text:
            str, the Arxiv identifier to parse.
    """
    # Check if the identifier starts with the url prefix, if so remove
    if id_text.startswith(ID_PREFIX):
        id_text = id_text[len(ID_PREFIX) :]

    # Check if the identifier is in the new format
    match = re.match(ID_PATTERN_NEW, id_text)
    if match:
        return Identifier(
            number=match.group("number"),
            version=match.group("version"),
            is_new=True,
        )

    # Check if the identifier is in the old format
    match = re.match(ID_PATTERN_OLD, id_text)
    if match:
        return Identifier(
            number=match.group("number"),
            version=None,
            is_new=False,
        )

    # If the identifier does not match either pattern, raise an error
    raise ValueError(f"Invalid Arxiv identifier: {id_text}")


@dataclass
class Entry:
    """A dataclass for an entry from the Arxiv API [4]

    Args:
        title:
            str, the title of the entry
        id:
            str, the Arxiv ID of the entry
        published:
            datetime.datetime, the published date of the entry
        updated:
            datetime.datetime, the updated date of the entry
        summary:
            str, the summary of the entry
        author:
            List[Author], the authors of the entry
        category:
            List[Category], the categories of the entry
    """

    title: str
    id: Identifier
    published: datetime.datetime
    updated: datetime.datetime
    summary: str
    author: List[Author]
    category: List[Category]

    def link(self) -> str:
        """Formatted arxiv link for the entry"""
        return f"https://arxiv.org/abs/{self.id}"

    def to_dict(self) -> dict:
        """Return the entry as a dictionary."""
        return {
            FIELD_ENTRY_TITLE: self.title,
            FIELD_ENTRY_ID: self.id,
            FIELD_ENTRY_PUBLISHED: self.published,
            FIELD_ENTRY_UPDATED: self.updated,
            FIELD_ENTRY_SUMMARY: self.summary,
            FIELD_ENTRY_AUTHOR: self.author,
            FIELD_ENTRY_CATEGORY: self.category,
        }

    @staticmethod
    def from_dict(data: dict) -> "Entry":
        """Create an entry from a dictionary."""
        return Entry(
            title=data[FIELD_ENTRY_TITLE],
            id=data[FIELD_ENTRY_ID],
            published=data[FIELD_ENTRY_PUBLISHED],
            updated=data[FIELD_ENTRY_UPDATED],
            summary=data[FIELD_ENTRY_SUMMARY],
            author=data[FIELD_ENTRY_AUTHOR],
            category=data[FIELD_ENTRY_CATEGORY],
        )

    def filter_authors(self, authors: List[str]) -> bool:
        """Check if the entry has any of the given authors.

        Args:
            authors:
                List[str], the list of authors to check for.

        Returns:
            bool: True if the entry has any of the authors, False otherwise.
        """
        return any(author.name in authors for author in self.author)


class SortBy(str, enum.Enum):
    """Enumeration of sort fields for the Arxiv API [3]"""

    Relevance = "relevance"
    LastUpdatedDate = "lastUpdatedDate"
    SubmittedDate = "submittedDate"


class SortOrder(str, enum.Enum):
    """Enumeration of sort orders for the Arxiv API [3]"""

    Ascending = "ascending"
    Descending = "descending"


class SearchField:
    """Enumeration of search fields for the Arxiv API [3]"""

    TITLE = "ti"
    AUTHOR = "au"
    ABSTRACT = "abs"
    COMMENT = "co"
    JOURNAL_REFERENCE = "jr"
    CATEGORY = "cat"
    REPORT_NUMBER = "rn"
    ID = "id"
    ALL = "all"


class LogicalOperator:
    """Enumeration of logical operators for the Arxiv API [3]"""

    AND = " AND "
    OR = " OR "
    AND_NOT = " ANDNOT "


def find(
    entry: ElementTree.Element, tag: str, find_all: bool = False
) -> Union[str, List[str]]:
    """Find the tag in the entry and return the text.

    Args:
        entry (ElementTree.Element): The entry to search.
        tag (str): The tag to search for.

    Returns:
        str: The text of the tag.
    """
    if not tag.startswith(TAG_PREFIX):
        tag = TAG_PREFIX + tag

    if find_all:
        return entry.findall(tag)

    res = entry.find(tag)
    if res is not None:
        return res.text


def parse_entry(entry: ElementTree.Element) -> Entry:
    """Parse an entry from the Arxiv API response.

    Args:
        entry (ElementTree.Element): The entry to parse.

    Returns:
        Entry: The parsed entry.
    """
    # Parse the authors
    authors = [
        Author(name=find(author, TAG_NAME), affiliation=find(author, TAG_AFFILIATION))
        for author in find(entry, TAG_AUTHOR, find_all=True)
    ]

    # Parse the categories
    categories = [
        Category(term=category.attrib[TAG_TERM], scheme=category.attrib[TAG_SCHEME])
        for category in find(entry, TAG_CATEGORY, find_all=True)
    ]

    # Parse date-type fields
    published = find(entry, TAG_PUBLISHED)
    updated = find(entry, TAG_UPDATED)
    if published is not None:
        published = datetime.datetime.fromisoformat(published)
    if updated is not None:
        updated = datetime.datetime.fromisoformat(updated)

    # Parse components of the identifier
    raw_id = find(entry, TAG_ID)
    id_ = parse_identifier(raw_id)

    # Return the parsed entry
    return Entry(
        title=find(entry, TAG_TITLE),
        id=id_,
        published=published,
        updated=updated,
        summary=find(entry, TAG_SUMMARY),
        author=authors,
        category=categories,
    )


def get_and_parse(url: str, params: dict) -> List[Entry]:
    """Get and parse the response from the Arxiv API, the payloads
    are encoded using the Atom 1 XML format.

    Args:
        url (str): The endpoint to query
        params (dict): The parameters to pass to the query

    Returns:
        dict: The parsed response
    """
    # Get the response
    response = SESSION.get(url, params=params)

    # Parse the response securely into ElementTree
    root = SecureElementTree.fromstring(response.text)

    # TODO finish parsing response into a list of named tuples if no errors,
    #  otherwise raise the error
    if len(root) == 1 and root[0].tag == "error":
        raise ValueError(f"No results found. Error: {root[0].text}")

    entries = [parse_entry(entry) for entry in find(root, TAG_ENTRY, find_all=True)]

    # Return the parsed response
    return entries


def _extend_query(
    query: str,
    field: SearchField,
    value: Union[str, List[str]],
    how: LogicalOperator = LogicalOperator.AND,
    how_list: LogicalOperator = LogicalOperator.OR,
) -> str:
    """Extend the query with the given field and value.

    Args:
        query:
            str, The current query string.
        field:
            SearchField, The field to search in.
        value:
            Union[str, List[str]], The value to search for.
        how:
            LogicalOperator, The logical operator to use when adding the field.
        how_list:
            LogicalOperator, The logical operator to use when adding a list of values.

    Returns:
        str: The extended query string.
    """
    # Check if query exists, if so then extend with cross-field logical operator
    if query:
        query += how

    # Check if value is scalar or list, then extend query
    if isinstance(value, str):
        query += f"{field}:{value}"
    elif isinstance(value, list):
        query += f"({field}:{how_list.join(value)})"

    return query


def format_search_query(
    title: Union[str, List[str]] = None,
    author: Union[str, List[str]] = None,
    abstract: Union[str, List[str]] = None,
    comment: Union[str, List[str]] = None,
    journal_reference: Union[str, List[str]] = None,
    category: Union[str, List[str]] = None,
    report_number: Union[str, List[str]] = None,
    id_list: List[str] = None,
    all_: Union[str, List[str]] = None,
    how: LogicalOperator = LogicalOperator.AND,
    how_list: LogicalOperator = LogicalOperator.OR,
) -> Union[str, None]:
    """Format the search query for the Arxiv API.

    Args:
        title:
            Union[str, List[str]], optional, The title to search for. Defaults to None.
        author:
            Union[str, List[str]], optional, The author to search for. Defaults to None.
        abstract:
            Union[str, List[str]], optional, The abstract to search for. Defaults to
            None.
        comment:
            Union[str, List[str]], optional, The comment to search for. Defaults to
            None.
        journal_reference:
            Union[str, List[str]], optional, The journal reference to search for.
            Defaults to None.
        category:
            Union[str, List[str]], optional, The category to search for. Defaults to
            None.
        report_number:
            Union[str, List[str]], optional, The report number to search for.
            Defaults to None.
        id_list:
            List[str], optional, The list of Arxiv IDs to search for. Defaults to None.
        all_:
            Union[str, List[str]], optional, The all field to search for. Defaults to
            None.
        how:
            LogicalOperator, optional, The logical operator to use when adding the
            field. Defaults to LogicalOperator.AND.
        how_list:
            LogicalOperator, optional, The logical operator to use when adding a list
            of values. Defaults to LogicalOperator.OR.

    Returns:
        str or None: The formatted query string, or None if no fields are provided.
    """
    # Short-circuit if no fields are provided
    if all(
        v is None
        for v in (
            title,
            author,
            abstract,
            comment,
            journal_reference,
            category,
            report_number,
            id_list,
            all_,
        )
    ):
        return None

    query = ""

    for field, value in zip(
        (
            SearchField.TITLE,
            SearchField.AUTHOR,
            SearchField.ABSTRACT,
            SearchField.COMMENT,
            SearchField.JOURNAL_REFERENCE,
            SearchField.CATEGORY,
            SearchField.REPORT_NUMBER,
            SearchField.ID,
            SearchField.ALL,
        ),
        (
            title,
            author,
            abstract,
            comment,
            journal_reference,
            category,
            report_number,
            id_list,
            all_,
        ),
    ):
        if value is not None:
            query = _extend_query(query, field, value, how=how, how_list=how_list)

    return parse.quote(query, safe="/:&=")


def _query(
    search_query: str = None,
    id_list: List[str] = None,
    start: int = 0,
    max_results: int = 10,
    sort_by: SortBy = SortBy.Relevance,
    sort_order: SortOrder = SortOrder.Descending,
    min_date: datetime.datetime = None,
    max_date: datetime.datetime = None,
    date_filter_field: str = FIELD_ENTRY_UPDATED,

) -> List[Entry]:
    """Query the Arxiv API with the given parameters.

    Args:
        search_query:
            str, optional, The query string to search for. Defaults to None.
        id_list:
            List[str], optional, A list of Arxiv IDs to search for. Defaults to None.
        start:
            int, optional, The index to start the search from. Defaults to 0.
        max_results:
            int, optional, The maximum number of results to return. Defaults to 10.
        sort_by:
            SortBy, optional, The field to sort by. Defaults to SortBy.Relevance.
        sort_order:
            SortOrder, optional, The order to sort by. Defaults to SortOrder.Descending.

    Returns:
        List[Entry]: The list of entries returned by the query.
    """
    # Define the parameters for the query
    params = {
        "search_query": search_query,
        "id_list": id_list,
        "start": start,
        "max_results": max_results,
        "sortBy": sort_by.value,
        "sortOrder": sort_order.value,
    }

    # Filter out the None values
    params = {k: v for k, v in params.items() if v is not None}

    # Get and parse the response
    results = get_and_parse(URL_QUERY, params)

    # Filter for dates if specified
    if date_filter_field not in (FIELD_ENTRY_PUBLISHED, FIELD_ENTRY_UPDATED):
        raise ValueError(
            f"Invalid date filter field: {date_filter_field}, options "
            f"are {FIELD_ENTRY_PUBLISHED} or {FIELD_ENTRY_UPDATED}"
        )
    if min_date is not None:
        results = [r for r in results if getattr(r, date_filter_field) >= min_date]
    if max_date is not None:
        results = [r for r in results if getattr(r, date_filter_field) <= max_date]

    # Return the results
    return results


def query(
    title: Union[str, List[str]] = None,
    author: Union[str, List[str]] = None,
    abstract: Union[str, List[str]] = None,
    comment: Union[str, List[str]] = None,
    journal_reference: Union[str, List[str]] = None,
    category: Union[str, List[str]] = None,
    report_number: Union[str, List[str]] = None,
    id_list: List[str] = None,
    all_: Union[str, List[str]] = None,
    how: LogicalOperator = LogicalOperator.AND,
    how_list: LogicalOperator = LogicalOperator.OR,
    start: int = 0,
    max_results: int = 10,
    sort_by: SortBy = SortBy.Relevance,
    sort_order: SortOrder = SortOrder.Descending,
    min_date: datetime.datetime = None,
    max_date: datetime.datetime = None,
    date_filter_field: str = FIELD_ENTRY_UPDATED,
) -> List[Entry]:
    """Query the Arxiv API with the given parameters.

    Args:
        title:
            Union[str, List[str]], optional, The title to search for. Defaults to None.
        author:
            Union[str, List[str]], optional, The author to search for. Defaults to None.
        abstract:
            Union[str, List[str]], optional, The abstract to search for. Defaults to
            None.
        comment:
            Union[str, List[str]], optional, The comment to search for. Defaults to
            None.
        journal_reference:
            Union[str, List[str]], optional, The journal reference to search for.
            Defaults to None.
        category:
            Union[str, List[str]], optional, The category to search for. Defaults to
            None.
        report_number:
            Union[str, List[str]], optional, The report number to search for.
            Defaults to None.
        id_list:
            List[str], optional, The list of Arxiv IDs to search for. Defaults to None.
        all_:
            Union[str, List[str]], optional, The all field to search for. Defaults to
            None.
        how:
            LogicalOperator, optional, The logical operator to use when adding the
            field. Defaults to LogicalOperator.AND.
        how_list:
            LogicalOperator, optional, The logical operator to use when adding a list
            of values. Defaults to LogicalOperator.OR.
        start:
            int, optional, The index to start the search from. Defaults to 0.
        max_results:
            int, optional, The maximum number of results to return. Defaults to 10.
        sort_by:
            SortBy, optional, The field to sort by. Defaults to SortBy.Relevance.
        sort_order:
            SortOrder, optional, The order to sort by. Defaults to SortOrder.Descending.
    """
    # Format the search query
    search_query = format_search_query(
        title,
        author,
        abstract,
        comment,
        journal_reference,
        category,
        report_number,
        id_list,
        all_,
        how,
        how_list,
    )

    # Short-circuit if no search query is provided
    if search_query is None and id_list is None:
        raise ValueError("No search query provided; cannot query the entire Arxiv.")

    # Query the API
    return _query(
        search_query=search_query,
        id_list=id_list,
        start=start,
        max_results=max_results,
        sort_by=sort_by,
        sort_order=sort_order,
        min_date=min_date,
        max_date=max_date,
        date_filter_field=date_filter_field,
    )



"""Utilities for sending emails using Gmail.
"""

import os
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

EMAIL_SENDER = "chalkdust@sxolar.org"
EMAIL_APP_PASSWORD_ENV_KEY = "SXOLAR_EMAIL_APP_PASSWORD"

try:
    EMAIL_APP_PASSWORD = os.environ[EMAIL_APP_PASSWORD_ENV_KEY]
except KeyError:
    EMAIL_APP_PASSWORD = None


def send_email(
    subject: str,
    to: list[str],
    body: str,
    from_email: str = EMAIL_SENDER,
    safe: bool = True,
    is_plain: bool = True,
    app_password: str = EMAIL_APP_PASSWORD,
):
    """Send an email using Gmail.

    Args:
        subject:
            str, The subject of the email.
        to:
            str, The email address to send the email to.
        body:
            str, The body of the email.
        from_email:
            str, default EMAIL_SENDER, the email address to send the email from.
        safe:
            bool, default True, whether to suppress exceptions when sending the email.
        is_plain:
            bool, default True, whether the email is plain text or html.
        app_password:
            str, default EMAIL_APP_PASSWORD, the app password for the Gmail account.
    """
    if app_password is None:
        raise ValueError(
            f"Please set the {EMAIL_APP_PASSWORD_ENV_KEY} environment variable to "
            f"your Gmail app password or specify it as the app_password argument."
        )

    # Create a multipart message and set headers
    message = MIMEMultipart("alternative")
    message["Subject"] = subject
    message["From"] = EMAIL_SENDER
    message["To"] = ", ".join(to)

    # Check if the email is plain text or html
    if is_plain:
        part = MIMEText(body, "plain")
        message.attach(part)
    else:
        html_content = f"""\
        <html>
          <body>
            {body}
          </body>
        </html>
        """

        # Turn these into MIMEText objects and attach them to the MIMEMultipart message
        part = MIMEText(html_content, "html")
        message.attach(part)

    # Connect to Gmail's SMTP server and send the email
    try:
        # Gmail SMTP server details
        smtp_server = "smtp.gmail.com"
        smtp_port = 587  # TLS port

        # Establish a secure session using starttls
        with smtplib.SMTP(smtp_server, smtp_port) as server:
            server.starttls()
            server.login(from_email, app_password)
            server.send_message(message)

    except Exception as e:
        if safe:
            print(f"Error sending email: {e}")
        else:
            raise e



"""Utilities for summarizing the results of arXiv queries.
"""

import datetime
import enum
from dataclasses import dataclass
from typing import List, Optional

from sxolar.api.arxiv import Entry, LogicalOperator, SortBy, SortOrder
from sxolar.api.search import Query


class Format(str, enum.Enum):
    """An enumeration of formatting options for summaries"""

    Plain = "plain_text"
    Html = "html"


@dataclass
class Section:
    """A section of the results of an arXiv query, with optional
    formatting behavior for plain text or html. A summary of results
    is essentially a list of entries with optional formatting options.

    Args:
        name:
            str, the name of the summary
        results:
            List[Entry], the list of entries to summarize
        max_authors:
            int, default 3, the maximum number of authors to include in the summary.
            Set to None to include all authors.
        include_abstract:
            bool, default False, whether to include the abstract in the summary
        max_results:
            int, default 50, the maximum number of results to include in the summary
        trailing:
            Optional[int], the number of days to include in the summary. If None,
            include all results.
        trailing_unit:
            str, default "days", the unit of time for the trailing period, must be a
            valid argument for datetime.timedelta
    """

    name: str
    query: Query
    results: List[Entry] = None
    max_authors: int = 3
    include_abstract: bool = False
    max_results: int = 50
    sort_by: SortBy = SortBy.Relevance
    sort_order: SortOrder = SortOrder.Descending
    trailing: Optional[int] = None
    trailing_unit: str = "days"

    def _format_entry(
        self,
        entry: Entry,
        format: Format,
        starting_header: int = 3,
    ) -> str:
        """Format an entry as a string"""
        authors = (
            entry.author
            if self.max_authors is None
            else entry.author[: self.max_authors]
        )
        authors = ", ".join(str(a) for a in authors)
        title = entry.title
        abstract = entry.summary if self.include_abstract else ""

        if format == Format.Plain:
            return (
                f"{title} [{entry.id}]\n"
                f"{authors}\n"
                f"{abstract}\n"
                f"{entry.id.link()}\n"
            )

        if format == Format.Html:
            return (
                f"<p>"
                f'<h{starting_header}><a href="{entry.id.link()}">{title} '
                f"[{entry.id}]</a></h{starting_header}><br>"
                f"{authors}<br>"
                f"{abstract}<br>"
                f"</p>"
            )

        raise ValueError(f"Invalid format: {format}, options are {Format}")

    def refresh(self):
        """Refresh the results of the query"""
        kwargs = {
            "start": 0,
            "max_results": self.max_results,
            "sort_by": self.sort_by,
            "sort_order": self.sort_order,
        }

        # Add time filters if specified
        if self.trailing is not None:
            # Format timedelta
            td = datetime.timedelta(**{self.trailing_unit: self.trailing})
            # Get the current date
            now = datetime.datetime.now(datetime.UTC)
            # Subtract the timedelta
            start_date = now - td
            kwargs["min_date"] = start_date
            kwargs["max_date"] = now

        # run the search
        self.results = self.query.search(**kwargs)

    def to_text(self) -> str:
        """Returns the summary as plain text"""
        return f"{self.name}:\n" + "\n".join(
            self._format_entry(entry, Format.Plain) for entry in self.results
        )

    def to_html(self, starting_header: int = 2) -> str:
        """Returns the summary as html"""
        return f"<h{starting_header}>{self.name}</h{starting_header}>\n" + "\n".join(
            self._format_entry(entry, Format.Html, starting_header=starting_header + 1)
            for entry in self.results
        )

    @staticmethod
    def from_combo(
        name: str,
        authors: List[str] = None,
        alls: List[str] = None,
        max_authors: int = 3,
        include_abstract: bool = False,
        max_results: int = 50,
        trailing: Optional[int] = None,
        trailing_unit: str = "days",
        filter_authors: bool = False,
        alls_operator: str = LogicalOperator.OR,
        value: str = None,
    ) -> "Section":
        """Create a section from a combination of arguments

        Args:

        """
        # Check for fully formed query
        if value is not None:
            query = Query(value, filter_authors=authors if filter_authors else None)

        # Check for authors or alls
        else:
            if authors is None and alls is None:
                raise ValueError("At least one of authors or alls must be specified")

            query = Query.from_combo(
                authors=authors,
                alls=alls,
                filter_authors=filter_authors,
                alls_operator=alls_operator,
            )

        # Check if trailing is nested dict
        if isinstance(trailing, dict):
            trailing_unit = trailing.get("unit", "days")
            trailing = trailing.get("num", None)

        return Section(
            name=name,
            query=query,
            max_authors=max_authors,
            include_abstract=include_abstract,
            max_results=max_results,
            trailing=trailing,
            trailing_unit=trailing_unit,
        )


@dataclass
class Summary:
    """A summary is a named collection of sections, with formatting metadata

    Args:
        name:
            str, the name of the summary
        sections:
            List[Section], a list of sections to include in the summary
    """

    name: str
    sections: List[Section]

    def refresh(self):
        """Refresh the results of the query"""
        for section in self.sections:
            section.refresh()

    def to_text(self) -> str:
        """Returns the summary as plain text"""
        return "\n\n".join(section.to_text() for section in self.sections)

    def to_html(self) -> str:
        """Returns the summary as html"""
        return "<br><br>".join(section.to_html() for section in self.sections)



"""Tests for search api
"""

from sxolar.api.search import Author, Title, Query, Abstract, All, JournalRef, Category


class TestQuery:
    """Test the Query base class"""

    def test_create(self):
        """Test the creation of a query"""
        q = Query('au:del maestro')
        assert isinstance(q, Query)
        assert q.value == 'au:del maestro'

    def test_str(self):
        """Test the string representation"""
        q = Query('au:del maestro')
        assert str(q) == 'au:del maestro'

    def test_and(self):
        """Test the and operator"""
        q = Query('au:del maestro')
        q2 = Query('ti:checkerboard')
        q3 = q & q2
        assert str(q3) == 'au:del maestro AND ti:checkerboard'

    def test_or(self):
        """Test the or operator"""
        q = Query('au:del maestro')
        q2 = Query('ti:checkerboard')
        q3 = q | q2
        assert str(q3) == 'au:del maestro OR ti:checkerboard'

    def test_and_not(self):
        """Test the and_not operator"""
        q = Query('au:del maestro')
        q2 = Query('ti:checkerboard')
        q3 = q - q2
        assert str(q3) == 'au:del maestro ANDNOT ti:checkerboard'

    def test_wrap(self):
        """Test the wrap function"""
        q = Query('au:del maestro')
        q2 = Query('ti:checkerboard')
        q3 = q & q2
        q4 = q3.wrap()
        assert str(q4) == '(au:del maestro AND ti:checkerboard)'

    def test_search(self):
        """Test the query function"""
        res = Author('del maestro').search()
        assert len(res) == 10

    def test_join(self):
        """Test the join function"""
        res = Author('del maestro').join(Title('checkerboard'))
        assert str(res) == '(au:del maestro OR ti:checkerboard)'


class TestSearchFieldClasses:
    """Test the search field classes"""

    def test_author(self):
        """Test the author class"""
        q = Author('del maestro')
        assert isinstance(q, Author)
        assert q.value == 'au:del maestro'

    def test_title(self):
        """Test the title class"""
        q = Title('checkerboard')
        assert isinstance(q, Title)
        assert q.value == 'ti:checkerboard'

    def test_abstract(self):
        """Test the abstract class"""
        q = Abstract('checkerboard')
        assert isinstance(q, Abstract)
        assert q.value == 'abs:checkerboard'

    def test_all(self):
        """Test the all class"""
        q = All('checkerboard')
        assert isinstance(q, All)
        assert q.value == 'all:checkerboard'

    def test_journal_ref(self):
        """Test the journal_ref class"""
        q = JournalRef('checkerboard')
        assert isinstance(q, JournalRef)
        assert q.value == 'jr:checkerboard'

    def test_category(self):
        """Test the category class"""
        q = Category('checkerboard')
        assert isinstance(q, Category)
        assert q.value == 'cat:checkerboard'
